import config
import util

from boto.vpc import VPCConnection

_cfg = config.Config()

# class method
def list():
    clusters = _cfg.get('clusters')
    cluster_names = clusters.keys()
    cluster_names.sort()
    print "Defined clusters:"
    for cluster_name in cluster_names:
        cluster = Cluster(cluster_name)
        status = cluster.status()
        print "    {}:".format(cluster_name)
        for k,v in status.items():
            print "        {}: {}".format(k, v)

def vpc_conn():
    args = util.connection_args()
    return VPCConnection(**args)

def get_vpc_by_name(name):
    vpcs = vpc_conn().get_all_vpcs()
    ret = None
    for vpc in vpcs:
        tags = vpc.tags
        if 'Name' not in tags:
            continue
        vpc_name = tags.get('Name')
        if vpc_name == name:
            # duplicate?
            if ret:
                print "Warning: found more than one VPC with the name {}".format(name)
            ret = vpc
    return ret

class Cluster:
    def __init__(self, name):    
        if name not in _cfg.get('clusters'):
            print "Invalid cluster name: {}".format(name)
            return
        self.cluster_config = _cfg.get('clusters', name)
        self.name = name
        self.conn = vpc_conn()

    def status(self):
        vpc = get_vpc_by_name(self.name)
        if not vpc:
            return { 'exists': False }

        # check if this VPC is managed by udo
        tags = vpc.tags
        vpc_is_managed = False
        if 'Udo' in tags:
            vpc_is_managed = True

        return {
            'udo': vpc_is_managed,
            'exists': True,
            'state': vpc.state,
            'id': vpc.id
        }
    
    # active this cluster
    def activate(self):
        conn = vpc_conn()

        cfg = _cfg.get('clusters', self.name)
        if not cfg:
            print "No configuration found for {}".format(self.name)
            return False

        vpc = get_vpc_by_name(self.name)
        if vpc:
            print "Cluster {} already exists".format(self.name)
            return

        # create VPC
        subnet_cidr = cfg.get('subnet_cidr')
        if not subnet_cidr:
            print "No subnet definition found for {}".format(self.name)
            return False
        vpc = conn.create_vpc(subnet_cidr)
        vpc.add_tag('Name', value=self.name)
        # for now assume that our subnet is the same CIDR as the VPC
        # this is simpler but less fancy
        subnet = conn.create_subnet(vpc.id, subnet_cidr)
        # all done
        util.message_integrations("Created VPC {}".format(self.name))

        # mark that this cluster is being managed by udo
        vpc.add_tag('udo', value=True)

        return True






# create_deployment(application_name, deployment_group_name=None, 
# revision=None, deployment_config_name=None, description=None, 
# ignore_application_stop_failures=None)

import config
import util
import boto
from datetime import datetime
from boto.codedeploy.layer1 import CodeDeployConnection

_cfg = config.Config()

class Deploy:
    # role_name is totally optional
    # cluster_name is kinda optional
    def __init__(self, cluster_name=None, role_name=None):
        if cluster_name:
            self.cluster_name = cluster_name
            if role_name:
                self.role_name = role_name
                self.cfg = config.get_role_config(cluster_name, role_name)
            else:
                self.cfg = config.get_cluster_config(cluster_name)
        else:
            self.cfg = _cfg.get_root()
        self.conn = util.deploy_conn()

    def config(self):
        return self.cfg.get('deploy')

    def app_name(self):
        application = self.config().get('application')
        if not application:
            print "Deployment application not specified or configured"
            print "Valid applications are:"
            self.list_applications()
            return

        return application

    def commit_id_display(self, commit_id):
        return commit_id[:10]

    # create a deployment
    def create(self, group_name, commit_id):
        cfg = self.config()

        if not 'application' in cfg:
            print "deployment application not specified in deployment configuration"
            self.list_applications()
            return

        # get source info
        # assume github for now
        repo_name = None
        source = None
        rev_type = 'github'
        if not 'github' in cfg:
            print "github info not specified in deployment configuration"
            return
        source = cfg['github']
        if not 'repo' in source:
            print "deployment github repository not specified in deployment configuration"
            return
        repo_name = source['repo']

        application_name = cfg['application']
        deploy_rev = {
            'revisionType': 'GitHub',
            'gitHubLocation': {
                'repository': repo_name,
                'commitId': commit_id,
            }
        }
        msg = "Deploying commit {} to deployment group: {}".format(self.commit_id_display(commit_id), group_name)
        deployment = self.conn.create_deployment(application_name,
            deployment_group_name=group_name,
            revision=deploy_rev,
            ignore_application_stop_failures=False,
        )
        if not deployment:
            # prob won't reach here, will throw error instead
            print "Deployment failed"
            return
        deployment_id = deployment['deploymentId']
        util.message_integrations(msg)
        self.list_deployments(deployment_id)

    def list_deployments(self, dep_id=None, group=None):
        deps = self.conn.list_deployments(deployment_group_name=group)
        if dep_id:
            self.print_deployment(dep_id)
        else:
            for dep_id in deps['deployments']:
                self.print_deployment(dep_id)

    def print_last_deployment(self, **kwargs):
        application_name = self.app_name()
        if 'deployment_group_name' in kwargs and not 'application_name' in kwargs:
            kwargs['application_name'] = application_name
        deps = self.conn.list_deployments(**kwargs)['deployments']
        if not len(deps):
            print "No deployments found"
            return
        self.print_deployment(deps[0])

    def stop_deployment(self):
        deps = self.conn.list_deployments()['deployments']
        if not len(deps):
            print "No deployments found"
            return
        last_dep_id = deps[0]
        self.conn.stop_deployment(last_dep_id)
        print "Stopped {}".format(last_dep_id)
        self.print_last_deployment()

    def print_deployment(self, dep_id):
        dep = self.conn.get_deployment(dep_id)
        info = dep['deploymentInfo']
        dep_id = info['deploymentId']
        app_name = info['applicationName']
        group_name = info['deploymentGroupName']
        status = info['status']
        rev_info = info['revision']
        create_time = datetime.fromtimestamp(info['createTime'])
        create_time_display = create_time.strftime("%A, %d %B %Y %I:%M%p %Z")
        commit_id = 'unknown'
        #print dep
        msg = ""
        if 'errorInformation' in info:
            error_info = info['errorInformation']
            msg = error_info['message']
        # else get status...?

        if 'gitHubLocation' in rev_info:
            commit_id = self.commit_id_display(rev_info['gitHubLocation']['commitId'])
        print """ - {}/{} [{}]
     Created: {}
     Status: {}
     Message: {}
     Commit: {}
""".format(app_name, group_name, dep_id, create_time_display, status,
        msg, commit_id)

    def list_applications(self):
        apps = self.conn.list_applications()
        # TODO: fetch more apps via next_token if available
        app_names = apps['applications']
        for name in app_names:
            print " - Application: {}".format(name)

    def list_groups(self, application=None):
        application = self.app_name()
        groups = self.conn.list_deployment_groups(application)
        # TODO: fetch more groups via next_token if available
        group_names = groups['deploymentGroups']
        for name in group_names:
            group = self.conn.get_deployment_group(application, name)
            info = group['deploymentGroupInfo']
            style = info['deploymentConfigName']
            print " - Group: {}/{}  \t\t[{}]".format(application, name, style)
            # print target revision info
            if 'targetRevision' in info:
                target_rev = info['targetRevision']
                rev_type = target_rev['revisionType']
                if rev_type and rev_type == 'GitHub':
                    github_loc = target_rev['gitHubLocation']
                    print "      Repository: {}".format(github_loc['repository'])
                    if 'commitId' in github_loc:
                        print "      Last commit ID: {}".format(github_loc['commitId'])
            print ""

    def list_configs(self):
        cfgs = self.conn.list_deployment_configs()
        # TODO: fetch more cfgs via next_token if available
        cfg_names = cfgs['deploymentConfigsList']
        for name in cfg_names:
            print " - Configuration: {}".format(name)




import config
import util
import launchconfig
import cluster

from pprint import pprint

import boto
from boto.ec2.autoscale import AutoScaleConnection
from boto.ec2.autoscale import AutoScalingGroup
from boto.ec2.autoscale import ScalingPolicy
from boto.ec2.autoscale import Tag

_cfg = config.Config()

class AutoscaleGroup:
    def __init__(self, cluster_name, role_name):
        self.cluster_name = cluster_name
        self.role_name = role_name
        self.role_config = config.get_role_config(cluster_name, role_name)
        self.conn = util.as_conn()

    def name(self):
        return "-".join([self.cluster_name, self.role_name])

    def exists(self):
        asg = self.get_asgroup()
        if asg:
            return asg
        return False

    def find_vpc_subnet_by_cidr(self, cidr):
        vpc = cluster.get_vpc_by_name(self.cluster_name)
        if not vpc:
            print "Failed to find VPC {}".format(self.cluster_name)
            return None
        # look for subnet that matches
        subnets = cluster.vpc_conn().get_all_subnets(
            filters=[
                ('cidrBlock', cidr),
                ('vpcId', vpc.id),
            ],
        )
        if len(subnets):
            return subnets[0]
        print "Failed to find subnet {}".format(cidr)
        return None

    # returns LaunchConfig for this ASgroup
    # may or may not exist
    def lc(self):
        lc = launchconfig.LaunchConfig(self.cluster_name, self.role_name)
        if self.exists():
            asgroup = self.get_asgroup()
            if not asgroup:
                return None  # this might be a race condition between exists() and get_asgroup()
            blc = asgroup.launch_config_name
            if blc:
                lc.set_name(blc)
        return lc

    # returns true if the LC exists
    def activate_lc(self):
        # make sure we have a launchconfig activated
        lc = self.lc()
        if not lc:
            # might need to wait a min
            util.retry(lambda: self.lc(), 60)
            lc = self.lc()
        if not lc:
            print "Timed out waiting to create LaunchConfiguration"
            return false
        if lc.exists():
            print "Using LaunchConfig {}".format(lc.name())
            return True
        print "Creating LaunchConfig {}".format(lc.name())
        return lc.activate()

    # we can't modify a launchconfig in place, we have to create
    # a new one and set that as our lc
    def update_lc(self):
        oldlc = self.lc()
        # get new version
        lc = oldlc.update()
        # set lc
        asgroup = self.get_asgroup()
        lcname = lc.name()
        setattr(asgroup, 'launch_config_name', lcname)
        asgroup.update()
        # delete old
        conn = util.as_conn()
        if oldlc.name() is not lcname:
            util.retry(lambda: conn.delete_launch_configuration(oldlc.name()), 60)

    def get_asgroup(self):
        conn = util.as_conn()
        ags = conn.get_all_groups(names = [self.name()])
        if not len(ags):
            return None
        return ags[0]

    # get desired_size
    def get_scale_size(self):
        asgroup = self.get_asgroup()
        print "Desired: {}\nMin:{}\nMax:{}".format(asgroup.desired_capacity, asgroup.min_size, asgroup.max_size)

    # set desired_size
    def scale(self, desired):
        asgroup = self.get_asgroup()

        if desired < asgroup.min_size:
            print "Cannot scale: {} is lower than min_size ({})".format(desired, asgroup.min_size)
            return
        if desired > asgroup.max_size:
            print "Cannot scale: {} is greater than max_size ({})".format(desired, asgroup.max_size)
            if not util.confirm("Increase max_size to {}?".format(desired)):
                return
            asgroup.max_size = desired

        current = asgroup.desired_capacity
        asgroup.desired_capacity = desired
        asgroup.update()
        asgroup = self.get_asgroup()
        new = asgroup.desired_capacity
        if (new != current):
            msg = "Changed ASgroup {} desired_capacity from {} to {}".format(self.name(), current, new)
            util.message_integrations(msg)

    # kill of ASgroup and recreate it
    def reload(self):
        if not util.confirm("Are you sure you want to tear down the {} ASgroup and recreate it?".format(self.name())):
            return
        util.message_integrations("Reloading ASgroup {}".format(self.name()))
        self.deactivate()
        util.retry(lambda: self.activate(), 60)

    def deactivate(self):
        if not self.exists():
            return
        ag = self.get_asgroup()
        ag.min_size = 0
        ag.max_size = 0
        ag.desired_capacity = 0
        ag.update()
        ag.shutdown_instances()
        print "Deleting... this may take a few minutes..."
        if util.retry(lambda: ag.delete(), 500):
            util.message_integrations("Deleted ASgroup {}".format(self.name()))
            # delete launchconfig too
            lc = self.lc()
            lc.deactivate()
        else:
            util.message_integrations("Failed to delete ASgroup {}".format(self.name()))

    # creates the LaunchConfig
    def activate(self):
        conn = util.as_conn()
        cfg = self.role_config
        name = self.name()

        # ensure this LC already exists
        if not self.activate_lc():
            return False

        # look up subnet id
        subnets = [self.find_vpc_subnet_by_cidr(cidr) for cidr in cfg.get('subnets_cidr')]
        if not subnets or not len(subnets) or None in subnets:
            print "Valid subnets_cidr are required for {}/{}".format(self.cluster_name, self.role_name)
            return False
        print "Using subnets {}".format(", ".join([s.id for s in subnets]))
        print "AZs: {}".format(cfg.get('availability_zones'))

        # does the ASgroup already exist?
        ag = AutoScalingGroup(
            group_name=self.name(),
            load_balancers=cfg.get('elbs'),
            availability_zones=cfg.get('availability_zones'),
            vpc_zone_identifier=[s.id for s in subnets],
            launch_config=self.lc().name(),
            desired_capacity=cfg.get('scale_policy', 'desired'),
            min_size=cfg.get('scale_policy', 'min_size'),
            max_size=cfg.get('scale_policy', 'max_size'),
        )

        if not conn.create_auto_scaling_group(ag):
            print "Failed to create autoscale group"
            return False

        # prepare instance tags
        tags = cfg.get('tags')
        if not tags:
            tags = {}
        tags['cluster'] = self.cluster_name
        tags['role'] = self.role_name
        tags['Name'] = self.name()

        # apply tags        
        tag_set = [self.ag_tag(ag, k,v) for (k,v) in tags.iteritems()]
        conn.create_or_update_tags(tag_set)

        util.message_integrations("Activated ASgroup {}".format(name))

        return ag

    def ag_tag(self, ag, k, v):
        return Tag(
            key=k,
            value=v,
            propagate_at_launch=True,
            resource_id=ag.name
        )



# Handy common utilities for udo modules

import config
import boto
import json
import urllib2
from time import sleep
import sys
import os
import socket
import getpass

# don't use for tests
def default_config():
    return config.Config()

def region():
    return default_config().get('region')

# returns dict of arguments common to any boto connection we establish
def connection_args():
    _region = region()
    if not _region:
        # use the default region, I guess?
        print "Warning: default region is not configured"
        return "us-east-1"

    args = {
        'region': boto.ec2.get_region(_region)
    }
    return args

# autoscale connection
def as_conn():
    args = connection_args()
    region = args.pop('region')
    return boto.ec2.autoscale.connect_to_region(region.name, **args)

# codedeploy connection
def deploy_conn():
    args = connection_args()
    region = args.pop('region')
    return boto.codedeploy.connect_to_region(region.name, **args)

# ask a yes/no question
# returns true/false
def confirm(msg):
    yn = raw_input(msg + " (y/n) ")
    if yn.lower() == 'y':
        return True
    print "Aborted"
    return False

# keep trying proc until timeout or no exception is thrown
# use this when you're waiting for a change to take effect
# FIXME: actually respect timeout
def retry(proc, timeout):
    success = False
    ret = None
    while success == False:
        try:
            ret = proc()     
            success = True
        except boto.exception.BotoServerError as e:
            if default_config().get('debug'):
                # dump response
                print "Error: {}, retrying...".format(e)
            else:
                print('.'),
            sys.stdout.flush()
            sleep(5)
    print "\n"
    return ret

def user_and_host():
    username = getpass.getuser()
    hostname = socket.gethostname()
    return "{}@{}".format(username, hostname)

# also prints out msg
def message_integrations(msg):
    message_slack("[{}]  {}".format(user_and_host(), msg))
    print msg

def message_slack(msg):
    slack_cfg = default_config().new_root('slack')
    if not slack_cfg:
        # not configured
        return

    slack_url      = slack_cfg.get('url')
    slack_username = slack_cfg.get('username')
    slack_channel  = slack_cfg.get('channel')

    payload = {
        'username': slack_username,
        'text': msg,
        'channel': slack_channel,
    }

    data = json.dumps(payload)
    headers = {'Content-Type': 'application/json'}
    request = urllib2.Request(slack_url, data, headers=headers)
    return urllib2.urlopen(request).read()
